import functools
import json
from logging import getLogger
from typing import Any, Callable, Dict, List, Optional, TypeVar, Union, cast

from confluent_kafka import Message, Producer

from heizer._source.topic import HeizerTopic
from heizer.config import HeizerConfig

logger = getLogger(__name__)
PRODUCER_LIST = []

R = TypeVar("R")
F = TypeVar("F", bound=Callable[..., Any])


def _default_encoder(value: Union[Dict[Any, Any], str, bytes]) -> bytes:
    """
    :param value:
    :return: bytes

    Default Kafka message encoder, which will encode inputs to bytes

    """
    if isinstance(value, dict):
        return json.dumps(value).encode("utf-8")
    elif isinstance(value, str):
        return value.encode("utf-8")
    elif isinstance(value, bytes):
        return value
    else:
        raise ValueError(f"Input type is not supported: {type(value).__name__}")


class producer(object):
    # args
    config: HeizerConfig = HeizerConfig()
    topics: List[HeizerTopic]
    message_encoder: Callable[..., bytes] = _default_encoder
    call_back: Optional[Callable[..., Any]] = None
    key: Optional[str] = None
    headers: Optional[Dict[str, str]] = None

    # private properties
    _producer_instance: Optional[Producer] = None

    def __init__(
        self,
        topics: List[HeizerTopic],
        config: HeizerConfig = HeizerConfig(),
        message_encoder: Callable[..., bytes] = _default_encoder,
        call_back: Optional[Callable[..., Any]] = None,
        key: Optional[str] = None,
        headers: Optional[Dict[str, str]] = None,
    ):
        self.topics = topics
        self.config = config
        self.message_encoder = message_encoder
        self.call_back = call_back
        self.key = key
        self.headers = headers

        PRODUCER_LIST.append(self)

    @property
    def _producer(self) -> Producer:
        if not self._producer_instance:
            self._producer_instance = Producer(self.config.value)
        return self._producer_instance

    def __call__(self, func: F) -> F:
        @functools.wraps(func)
        def decorator(*args: Any, **kwargs: Any) -> Any:
            try:
                result = func(*args, **kwargs)
            except Exception as e:
                raise e

            msg = self.message_encoder(result)

            self._produce_message(message=msg)

            return result

        return cast(F, decorator)

    def _produce_message(self, message: bytes) -> None:
        for topic in self.topics:
            for partition in topic.partitions:
                try:
                    self._producer.poll(0)
                    self._producer.produce(
                        topic=topic.name,
                        value=message,
                        partition=partition,
                        key=self.key,
                        headers=self.headers,
                        on_delivery=self.call_back,
                    )
                    self._producer.flush()
                except Exception as e:
                    logger.error(f"Failed to produce msg. {str(e)}")
                    raise e


def delivery_report(err: str, msg: Message) -> None:
    """
    Called once for each message produced to indicate delivery result.
    Triggered by poll() or flush().

    :param msg:
    :param err:
    :return:
    """
    if err is not None:
        print("Message delivery failed: {}".format(err))
    else:
        print("Message delivered to {} [{}]".format(msg.topic(), msg.partition()))



import os

import pytest
from confluent_kafka.admin import AdminClient


@pytest.fixture()
def bootstrap_server() -> str:
    return os.environ.get("KAFKA_SERVER", "localhost:9092")


@pytest.fixture()
def admin_client(bootstrap_server) -> AdminClient:
    return AdminClient({"bootstrap.servers": bootstrap_server})



import logging
import os
import sys

FORMAT = "%(asctime)s %(levelname)-8s %(message)s"


def get_logger(name: str) -> logging.Logger:
    formatter = logging.Formatter(fmt=FORMAT, datefmt="%Y-%m-%d %H:%M:%S")
    handler = logging.FileHandler("/tmp/heizer_log.txt", mode="w")
    handler.setFormatter(formatter)
    screen_handler = logging.StreamHandler(stream=sys.stdout)
    screen_handler.setFormatter(formatter)
    logger = logging.getLogger(name)
    logger.setLevel(os.environ.get("HEIZER_LOG_LEVEL", "INFO"))
    logger.addHandler(handler)
    logger.addHandler(screen_handler)
    return logger



from typing import List, Optional

from confluent_kafka import TopicPartition


class HeizerTopic(TopicPartition):
    name: str
    _partitions: List[int]
    _topic_partitions: List[TopicPartition]

    def __init__(self, name: str, partitions: Optional[List[int]] = None):
        self._partitions = partitions or [-1]
        self._topic_partitions = []
        self.name = name

        for partition in self._partitions:
            self._topic_partitions.append(TopicPartition(topic=name, partition=partition))

    @property
    def partitions(self) -> List[int]:
        return self._partitions

    @property
    def topic_partitions(self) -> List[TopicPartition]:
        return self._topic_partitions



import asyncio
import functools
import logging
from typing import Any, Awaitable, Callable, Concatenate, Coroutine, List, Optional, ParamSpec, Type, TypeVar, Union
from uuid import uuid4

from confluent_kafka import Consumer
from pydantic import BaseModel

from heizer._source import get_logger
from heizer._source.message import HeizerMessage
from heizer._source.topic import HeizerTopic
from heizer.config import HeizerConfig
from heizer.types import Stopper

R = TypeVar("R")
F = TypeVar("F", bound=Callable[..., Any])
P = ParamSpec("P")
T = TypeVar("T")

logger = get_logger(__name__)


class consumer(object):
    """A decorator to create a consumer"""

    __id__: str

    topics: List[HeizerTopic]
    config: HeizerConfig = HeizerConfig()
    call_once: bool = False
    stopper: Optional[Stopper] = None
    deserializer: Optional[Type[BaseModel]] = None

    is_async: bool = False

    def __init__(
        self,
        *,
        topics: List[HeizerTopic],
        config: HeizerConfig = HeizerConfig(),
        call_once: bool = False,
        stopper: Optional[Stopper] = None,
        deserializer: Optional[Type[BaseModel]] = None,
        is_async: bool = False,
    ):
        self.topics = topics
        self.config = config
        self.call_once = call_once
        self.stopper = stopper
        self.deserializer = deserializer
        self.__id__ = str(uuid4())
        self.is_async = is_async

    async def __run__(
        self,
        func: Callable[Concatenate[HeizerMessage, P], Union[T, Awaitable[T]]],
        c: Consumer,
        is_async: bool,
        *args: P.args,
        **kwargs: P.kwargs,
    ) -> Union[Optional[T]]:  # ignore type
        """Run the consumer"""

        while True:
            msg = c.poll(1)

            if msg is None:
                continue

            if msg.error():
                logger.error(f"Consumer error: {msg.error()}")
                continue

            logger.debug("Received message")
            hmessage = HeizerMessage(msg)

            if self.deserializer is not None:
                logger.debug("Parsing message")
                try:
                    hmessage.formatted_value = self.deserializer.parse_raw(hmessage.value)
                except Exception as e:
                    logger.exception("Failed to deserialize message", exc_info=e)

            result = None

            logger.debug("Executing function")
            try:
                if is_async:
                    result = await func(hmessage, *args, **kwargs)  # type: ignore
                else:
                    result = func(hmessage, *args, **kwargs)
            except Exception as e:
                logger.exception(
                    f"Failed to execute function {func.__name__}",
                    exc_info=e,
                )
            finally:
                # TODO: add failed message to a retry queue
                logger.debug("Committing message")
                c.commit()

            if self.stopper is not None:
                logger.debug("Executing stopper function")
                try:
                    should_stop = self.stopper(hmessage)
                except Exception as e:
                    logger.warning(
                        f"Failed to execute stopper function {self.stopper.__name__}.",
                        exc_info=e,
                    )
                    should_stop = False

                if should_stop:
                    return result

            if self.call_once is True:
                logger.debug("Call Once is on, returning result")
                return result

    def __call__(
        self, func: Callable[Concatenate[HeizerMessage, P], T]
    ) -> Callable[P, Union[Coroutine[Any, Any, Optional[T]], T, None]]:
        logger.debug("Creating consumer")
        c = Consumer(self.config.value)

        logger.debug("Subscribing to topics")
        c.subscribe([topic.name for topic in self.topics])

        @functools.wraps(func)
        async def async_decorator(*args: P.args, **kwargs: P.kwargs) -> Optional[T]:
            """Async decorator"""
            logging.debug("Running async decorator")
            return await self.__run__(func, c, self.is_async, *args, **kwargs)

        @functools.wraps(func)
        def decorator(*args: P.args, **kwargs: P.kwargs) -> Optional[T]:
            """Sync decorator"""
            logging.debug("Running sync decorator")
            return asyncio.get_event_loop().run_until_complete(self.__run__(func, c, self.is_async, *args, **kwargs))

        return async_decorator if self.is_async else decorator



import asyncio

import websockets

from heizer import HeizerConfig, HeizerTopic, consumer

consumer_config = HeizerConfig(
    {
        "bootstrap.servers": "0.0.0.0:9092",
        "group.id": "test",
        "auto.offset.reset": "earliest",
    }
)

topics = [HeizerTopic(name="my.topic1")]


@consumer(topics=topics, config=consumer_config, is_async=True)
async def handler(message, websocket, *args, **kwargs):
    await websocket.send(message.value)


async def main():
    async with websockets.serve(handler, "", 8001):
        await asyncio.Future()


if __name__ == "__main__":
    asyncio.run(main())



# Configuration file for the Sphinx documentation builder.
#
# For the full list of built-in configuration values, see the documentation:
# https://www.sphinx-doc.org/en/master/usage/configuration.html

# -- Project information -----------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#project-information
# import os
# import sys

# sys.path.insert(0, os.path.abspath("../.."))

project = "heizer"
copyright = "2023, Yan Zhang"
author = "Yan Zhang"
release = "main"

# -- General configuration ---------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#general-configuration

extensions = [
    "myst_parser",
    "sphinx_multiversion",
    "sphinx.ext.autodoc",
    "sphinx.ext.viewcode",
    "sphinx.ext.napoleon",
    "IPython.sphinxext.ipython_directive",
    "IPython.sphinxext.ipython_console_highlighting",
]

templates_path = [
    "_templates",
]


autodoc_typehints = "description"


# -- Options for HTML output -------------------------------------------------
# https://www.sphinx-doc.org/en/master/usage/configuration.html#options-for-html-output

html_theme = "sphinx_rtd_theme"
html_static_path = ["_static"]

html_theme_options = {
    "analytics_id": "G-5RY57HDL43",
    "analytics_anonymize_ip": False,
    "display_version": False,
    "prev_next_buttons_location": "bottom",
    "style_external_links": False,
    "vcs_pageview_mode": "",
}


smv_tag_whitelist = r"^.*$"
smv_branch_whitelist = r"^.*$"
smv_remote_whitelist = r"^(origin|upstream)$"
smv_released_pattern = r"^tags/.*$"



from heizer._source.consumer import consumer
from heizer._source.message import HeizerMessage
from heizer._source.producer import producer
from heizer._source.topic import HeizerTopic
from heizer.config import HeizerConfig

__all__ = [
    "consumer",
    "producer",
    "HeizerConfig",
    "HeizerTopic",
    "HeizerMessage",
]



from typing import Any, List, Optional, Tuple

from confluent_kafka import Message


class HeizerMessage:
    # initialized properties
    message: Message
    topic: Optional[str]
    partition: int
    headers: Optional[dict[str, str]]
    key: Optional[str]
    value: Optional[str]

    formatted_value: Optional[Any] = None

    def __init__(self, message: Message):
        self.message = message
        self.topic = self._parse_topic(message.topic())
        self.partition = message.partition()
        self.headers = self._parse_headers(message.headers())
        self.key = self._parse_key(message.key())
        self.value = self._parse_value(message.value())

    @staticmethod
    def __bytes_to_str(b: bytes) -> str:
        """Convert bytes to string"""
        return str(b, "utf-8")

    def _parse_topic(self, topic: str | bytes) -> Optional[str]:
        """Parse topic to string"""
        if topic is None:
            return None
        return self.__bytes_to_str(topic) if isinstance(topic, bytes) else topic

    def _parse_headers(self, headers: Optional[List[Tuple[str, bytes]]]) -> Optional[dict[str, str]]:
        """Parse headers to dict"""
        parsed_headers = {}
        if headers is None:
            return None
        for k, v in headers:
            parsed_headers.update({k: v if isinstance(v, str) else self.__bytes_to_str(v)})
        return parsed_headers

    def _parse_key(self, key: str | bytes) -> Optional[str]:
        """Parse key to string"""
        if key is None:
            return None
        return self.__bytes_to_str(key) if isinstance(key, bytes) else key

    def _parse_value(self, value: str | bytes) -> Optional[str]:
        """Parse value to string"""
        if value is None:
            return None
        return self.__bytes_to_str(value) if isinstance(value, bytes) else value



import os
from typing import Optional

from heizer.types import KafkaConfig

DEFAULT_KAFKA_BOOTSTRAP_SERVER = os.environ.get("KAFKA_BOOTSTRAP_SERVER", "localhost:9094")
DEFAULT_KAFKA_GROUP = os.environ.get("KAFKA_GROUP", "default")


class HeizerConfig(object):
    # confluent kafka configs
    # https://docs.confluent.io/platform/current/clients/confluent-kafka-python/html/index.html#kafka-client-configuration
    __kafka_config: KafkaConfig

    def __init__(self, config: Optional[KafkaConfig] = None):
        self.__kafka_config = config or {
            "bootstrap.servers": DEFAULT_KAFKA_BOOTSTRAP_SERVER,
            "group.id": DEFAULT_KAFKA_GROUP,
        }

    @property
    def value(self) -> KafkaConfig:
        """
        Return kafka configurations

        Returns
        -------
        config : dict
            The dict contains kafka configurations
        """
        return self.__kafka_config



from pathlib import Path

from setuptools import setup

with open(Path(__file__).parent / "heizer" / "VERSION") as f:
    version = f.read().strip()

if __name__ == "__main__":
    setup(version=version)



import json
import logging
import os
from typing import Any, Dict, cast

from pydantic import BaseModel

from heizer import HeizerConfig, HeizerMessage, HeizerTopic, consumer, producer

Producer_Config = HeizerConfig(
    {
        "bootstrap.servers": os.environ.get("KAFKA_SERVER", "localhost:9092"),
    }
)

Consumer_Config = HeizerConfig(
    {
        "bootstrap.servers": os.environ.get("KAFKA_SERVER", "localhost:9092"),
        "group.id": "default",
        "auto.offset.reset": "earliest",
    }
)


def test_consumer_stopper() -> None:
    @producer(
        topics=[HeizerTopic(name="heizer.test.result")],
        config=Producer_Config,
    )
    def produce_data(status: str, result_value: str) -> Dict[str, Any]:
        return {
            "key1": 1,
            "key2": "2",
            "key3": True,
            "status": status,
            "result": result_value,
        }

    def stopper(msg: HeizerMessage) -> bool:
        data = json.loads(msg.value)
        if data["status"] == "success":
            return True
        return False

    @consumer(
        topics=[HeizerTopic(name="heizer.test.result")],
        config=Consumer_Config,
        stopper=stopper,
    )
    def consume_data(msg, *args, **kwargs) -> str:
        data = json.loads(msg.value)
        return cast(str, data["result"])

    produce_data("start", "waiting")
    produce_data("loading", "waiting")
    produce_data("success", "finished")
    produce_data("postprocess", "postprocess")

    result = consume_data()  # type: ignore

    assert result == "finished"


def test_consumer_call_once() -> None:
    topic_name = "heizer.test.test_consumer_call_once"

    @producer(
        topics=[HeizerTopic(name=topic_name)],
        config=Producer_Config,
    )
    def produce_data(status: str, result: str) -> Dict[str, Any]:
        return {
            "key1": 1,
            "key2": "2",
            "key3": True,
            "status": status,
            "result": result,
        }

    @consumer(
        topics=[HeizerTopic(name=topic_name)],
        config=Consumer_Config,
        call_once=True,
    )
    def consume_data(msg, *args, **kwargs) -> str:
        data = json.loads(msg.value)
        return cast(str, data["result"])

    produce_data("start", "waiting")
    produce_data("loading", "waiting")
    produce_data("success", "finished")

    result = consume_data()

    assert result == "waiting"


def test_consumer_deserializer(caplog) -> None:
    caplog.set_level(logging.DEBUG)
    topic_name = "heizer.test.test_consumer_deserializer"

    class TestModel(BaseModel):
        name: str
        age: int

    deserializer = TestModel

    @producer(
        topics=[HeizerTopic(name=topic_name)],
        config=Producer_Config,
    )
    def produce_data() -> Dict[str, Any]:
        return {
            "name": "mike",
            "age": 20,
        }

    @consumer(
        topics=[HeizerTopic(name=topic_name)],
        config=Consumer_Config,
        call_once=True,
        deserializer=deserializer,
    )
    def consume_data(message: HeizerMessage, *args, **kwargs):
        return message.formatted_value

    produce_data()

    result = consume_data()

    assert isinstance(result, TestModel)

    assert result.name == "mike"
    assert result.age == 20
